# encoding: utf-8

import re
from datetime import date
from numbers import Number

import odf.number
import odf.opendocument
import odf.style
import odf.table
import odf.text
from odf import grammar
from odf.namespaces import STYLENS, NUMBERNS


def _allowed_args(ns, element):
    return [
        x[1].lower().replace('-', '')
        for x in grammar.allowed_attributes[(ns, element)]
    ]


def _extract_args(args, allowed_args):
    return dict(
        (k, args.pop(k)) for k in allowed_args if k in args
    )


CELL_PROPERTIES_ARGS = _allowed_args(STYLENS, 'table-cell-properties')
COLUMN_PROPERTIES_ARGS = _allowed_args(STYLENS, 'table-column-properties')
ROW_PROPERTIES_ARGS = _allowed_args(STYLENS, 'table-row-properties')
TEXT_PROPERTIES_ARGS = _allowed_args(STYLENS, 'text-properties')
STYLE_ARGS = _allowed_args(STYLENS, 'style')
NUMBER_ARGS = _allowed_args(NUMBERNS, 'number')


class Wrapper(object):
    """
    Wrapper for odfpy constructors.

    Since odfpy constructors are really functions that all return instances
    of the Element class, we wrap the Element in order to provide additional
    methods.
    """

    def __init__(self, *args, **kwargs):
        self.__wrapped = self._wrapper_constructor(*args, **kwargs)

    def __getattr__(self, attr):
        return getattr(self.__wrapped, attr)


class Document(Wrapper):
    _wrapper_constructor = staticmethod(
        odf.opendocument.OpenDocumentSpreadsheet
    )

    def __init__(self, *args, **kwargs):
        super(Document, self).__init__(*args, **kwargs)
        self._automatic_style_idx = 0
        self._cell_styles = {}
        self._row_styles = {}
        self._column_styles = {}
        self._number_styles = {}
        self._percentage_styles = {}

    def add_table(self, **kwargs):
        table = Table(**kwargs)
        table._doc = self
        self.spreadsheet.addElement(table)
        return table

    def add_automatic_style(self, constructor, **kwargs):
        name = 'auto_style_%d' % self._automatic_style_idx
        self._automatic_style_idx += 1
        style = constructor(name=name, **kwargs)
        self.automaticstyles.addElement(style)
        return style

    def add_iso_date_style(self, name):
        style = odf.number.DateStyle(name=name)
        style.addElement(odf.number.Year(style='long'))
        style.addElement(odf.number.Text(text='-'))
        style.addElement(odf.number.Month(style='long'))
        style.addElement(odf.number.Text(text='-'))
        style.addElement(odf.number.Day(style='long'))
        self.automaticstyles.addElement(style)
        return style

    def cell_style_name(self, **kwargs):
        key = tuple(sorted(kwargs.items()))
        style = self._cell_styles.get(key)
        if style is None:
            style_args = _extract_args(kwargs, STYLE_ARGS)
            style = self.add_automatic_style(
                odf.style.Style, family='table-cell', **style_args
            )
            cell_args = _extract_args(kwargs, CELL_PROPERTIES_ARGS)
            if cell_args:
                style.addElement(odf.style.TableCellProperties(**cell_args))
            text_args = _extract_args(kwargs, TEXT_PROPERTIES_ARGS)
            if text_args:
                style.addElement(odf.style.TextProperties(**text_args))
            self._cell_styles[key] = style
        return style.getAttribute('name')

    def column_style_name(self, **kwargs):
        key = tuple(sorted(kwargs.items()))
        style = self._column_styles.get(key)
        if style is None:
            style_args = _extract_args(kwargs, STYLE_ARGS)
            style = self.add_automatic_style(
                odf.style.Style, family='table-column', **style_args
            )
            style.addElement(odf.style.TableColumnProperties(**kwargs))
            self._column_styles[key] = style
        return style.getAttribute('name')

    def row_style_name(self, **kwargs):
        key = tuple(sorted(kwargs.items()))
        style = self._row_styles.get(key)
        if style is None:
            style_args = _extract_args(kwargs, STYLE_ARGS)
            style = self.add_automatic_style(
                odf.style.Style, family='table-row', **style_args
            )
            style.addElement(odf.style.TableRowProperties(**kwargs))
            self._row_styles[key] = style
        return style.getAttribute('name')

    def number_style_name(self, **kwargs):
        key = tuple(sorted(kwargs.items()))
        style = self._number_styles.get(key)
        if style is None:
            style = self.add_automatic_style(odf.number.NumberStyle)
            style.addElement(odf.number.Number(**kwargs))
            self._number_styles[key] = style
        return style.getAttribute('name')

    def percentage_style_name(self, **kwargs):
        key = tuple(sorted(kwargs.items()))
        style = self._percentage_styles.get(key)
        if style is None:
            style = self.add_automatic_style(odf.number.PercentageStyle)
            style.addElement(odf.number.Number(**kwargs))
            style.addElement(odf.number.Text(text='%'))
            self._percentage_styles[key] = style
        return style.getAttribute('name')


class Table(Wrapper):
    _wrapper_constructor = staticmethod(odf.table.Table)

    def add_row(self, values=[], **kwargs):

        # Put row properties in an automatic style
        style_args = _extract_args(kwargs, ROW_PROPERTIES_ARGS + STYLE_ARGS)
        if style_args:
            kwargs['stylename'] = self._doc.row_style_name(**style_args)
        cell_attrs = _extract_args(
            kwargs, CELL_PROPERTIES_ARGS + TEXT_PROPERTIES_ARGS
        )
        row = Row(cell_attrs=cell_attrs, **kwargs)
        row._doc = self._doc
        row.add_cells(values)
        self.addElement(row)
        return row

    def add_column(self, **kwargs):

        # Put column properties in an automatic style
        style_args = _extract_args(kwargs, COLUMN_PROPERTIES_ARGS + STYLE_ARGS)
        if style_args:
            kwargs['stylename'] = self._doc.column_style_name(**style_args)

        col = odf.table.TableColumn(**kwargs)
        self.addElement(col)
        return col


class Row(Wrapper):
    _wrapper_constructor = staticmethod(odf.table.TableRow)

    def __init__(self, cell_attrs=None, **kwargs):
        super(Row, self).__init__(**kwargs)
        self._cell_attrs = cell_attrs

    def add_cell(self, value=None, **kwargs):

        # Combine row attrs and cell attrs
        if self._cell_attrs:
            attrs = self._cell_attrs.copy()
            attrs.update(kwargs)
        else:
            attrs = kwargs

        # Special treatment for numbers
        number_args = _extract_args(attrs, NUMBER_ARGS)
        if attrs.get('valuetype') == 'percentage':
            number_args.setdefault('minintegerdigits', 1)
            attrs['datastylename'] = \
                self._doc.percentage_style_name(**number_args)
        elif number_args:
            number_args.setdefault('minintegerdigits', 1)
            attrs['datastylename'] = \
                self._doc.number_style_name(**number_args)

        # Put cell properties in an automatic style
        style_args = _extract_args(
            attrs, CELL_PROPERTIES_ARGS + TEXT_PROPERTIES_ARGS + STYLE_ARGS
        )
        if style_args:
            attrs['stylename'] = self._doc.cell_style_name(**style_args)

        if value:
            if isinstance(value, Number):
                attrs.setdefault('valuetype', 'float')
                attrs['value'] = float(value)
                cell = odf.table.TableCell(**attrs)
            elif isinstance(value, date):
                attrs['valuetype'] = 'date'
                attrs['datevalue'] = value.isoformat()
                cell = odf.table.TableCell(**attrs)
            else:
                if not isinstance(value, basestring):
                    value = unicode(value)
                if '\r' in value or '\n' in value:
                    cell = odf.table.TableCell(**attrs)
                    lines = re.split('\r\n|\r|\n', value)
                    for line in lines:
                        p = odf.text.P()
                        p.addText(line)
                        cell.addElement(p)
                else:
                    attrs['stringvalue'] = value
                    cell = odf.table.TableCell(**attrs)
        else:
            cell = odf.table.TableCell(**attrs)

        self.addElement(cell)
        return cell

    def add_cells(self, values, **kwargs):
        for value in values:
            self.add_cell(value, **kwargs)
